/*
 *
 *	File: CmnLog.cpp 
 *  Desc: Functions and macros for diagnostic (TRACE, ...)
 *  Authors: MrAndrew, Swap
 *  Date: 23.12. 02
 *  Update:	31.12.2002
 *
 */
#include "StdAfx.h"

#define COMMON_API __declspec(dllexport)

#include "CmnError.h"
#include "CmnLog.h"
#include "CmnStr.h"
#include "CmnFile.h"

#include <dxerr8.h>

// The settings of the logger
LOGSETTINGS g_sLogSettings;

tagLogSettings::tagLogSettings()
{
	bUnicodeLogFile = TRUE;
	bLogToFile = TRUE;
	
	lpwszLogFile = NULL;
	hLogFile = NULL;
	dwFailLine = 0;
}

tagLogSettings::~tagLogSettings()
{
	// Delete dynamically allocated memory and close the file
	if(lpwszLogFile)
	{
		delete lpwszLogFile;
		lpwszLogFile = NULL;
	}

	// If the handle is still valid, then close it
	if(hLogFile)
	{
		cmnCloseFile(hLogFile);
		hLogFile = NULL;
	}
}

COMMON_API DWORD cmnInitLogW(LPCWSTR _lpwszFile, BOOL _bUnicodeLog)
{
	DWORD dwRet = ERROR_INVALID_PARAMETER;

	if(_lpwszFile)
	{
		WCHAR wszLogFilePath[MAX_PATH];

		g_sLogSettings.bLogToFile = TRUE;
		g_sLogSettings.bUnicodeLogFile = _bUnicodeLog;
			
		cmnGetDataFilePathW(wszLogFilePath, _lpwszFile);
		g_sLogSettings.hLogFile = cmnCreateFileW(wszLogFilePath);

		if(!g_sLogSettings.hLogFile)
		{
			return E_CMN_LOG_FILE_CREATE_FAIL;
		}

		dwRet = ERROR_SUCCESS;
	}

	return dwRet;
}

COMMON_API DWORD cmnInitLogA(LPCSTR _lpszFile, BOOL _bUnicodeLog)
{
	DWORD dwRet = ERROR_INVALID_PARAMETER;

	if(_lpszFile)
	{
		char szLogFilePath[MAX_PATH];

		g_sLogSettings.bLogToFile = TRUE;
		g_sLogSettings.bUnicodeLogFile = _bUnicodeLog;
			
		cmnGetDataFilePathA(szLogFilePath, _lpszFile);
		g_sLogSettings.hLogFile = cmnCreateFileA(szLogFilePath);

		if(!g_sLogSettings.hLogFile)
		{
			return E_CMN_LOG_FILE_CREATE_FAIL;
		}

		dwRet = ERROR_SUCCESS;
	}

	return dwRet;
}

// Returns log file name in coding, that the user needs
COMMON_API DWORD cmnGetLogFileNameW(LPWSTR _lpwszBuffer, LPDWORD _lpdwSize)
{
	DWORD dwRet = ERROR_INVALID_PARAMETER;

	if(_lpdwSize)
	{
		// If the log file name doesn't exist, then the log wasn't successfully initialized
		if(g_sLogSettings.wszFailFileName)
		{
			DWORD dwRealSize = (DWORD)(wcslen(g_sLogSettings.lpwszLogFile) + sizeof(WCHAR));	// the null char

			if((_lpwszBuffer) && (dwRealSize <= *_lpdwSize))
			{
				// The buffer exists and is big enough to hold the string
				wcscpy(_lpwszBuffer, g_sLogSettings.wszFailFileName);
			}
			else
			{
				// The user just wants to know how the buffer should be big
				*_lpdwSize = dwRealSize;
			}
		}
		else
		{
			return E_CMN_LOG_NOT_INIT;
		}
	}

	return dwRet;
}

COMMON_API DWORD cmnGetLogFileNameA(LPSTR _lpszBuffer, LPDWORD _lpdwSize)
{
	DWORD dwRet = ERROR_CALL_NOT_IMPLEMENTED;

	return dwRet;
}

/*
 *	Writes formatted string to output window and to log file
 *
 */
void cmnTraceW(LPCWSTR lpcwszFormat, ...)
{		
	DWORD w;
	int nBufferSize;
	WCHAR wszBuffer[512];	
	WCHAR wszOut[512];
	LPCWSTR lpFormat;
	BOOL	bPureMsg = FALSE;
	WCHAR	wszFileName[MAX_PATH];
	
	va_list args;
	va_start(args, lpcwszFormat);	

	WCHAR* cSemicolon = wcschr(lpcwszFormat, _S_SEMICOLON_CHARW);
	if(cSemicolon != NULL && cSemicolon == lpcwszFormat) {
		lpFormat = L"%s\n";
		bPureMsg = TRUE;
	}
	else {
		lpFormat = L"* %s[%d]: %s\n";
		bPureMsg = FALSE;
	}

	nBufferSize = _vsnwprintf(wszBuffer, _COUNTOF(wszBuffer), lpcwszFormat, args);
	va_end(args);

	if(nBufferSize >= 0) {
		
		if(bPureMsg) {
			nBufferSize = swprintf(wszOut, lpFormat, wszBuffer);
		}
		else {
			cmnCutDirectoryW(wszFileName, cmnGetFailFileName());
			nBufferSize = swprintf(wszOut, lpFormat, wszFileName, cmnGetFailLineNumber(), wszBuffer);
		}
		
		if(nBufferSize >= 0) {
			OutputDebugStringW(wszOut);
		}
	}
			
	if(cmnIsLoggingToFile() && cmnGetLogFileHandle())
	{
		if(cmnIsLogFileUnicode())
		{
			WCHAR c;

			for(int i = 0; i < (int)wcslen(wszOut); i++)
			{
				//
				// Get char from output buffer
				c = wszOut[i];
			
				//
				// Detects CRLF chars
				if(c == _S_NEWLINE_CHARW) {
					
					// Write CR
					c = _S_CR_CHARW;
					WriteFile(cmnGetLogFileHandle(), &c, sizeof(c), (LPDWORD)&w, NULL);
					
					// Write LF
					c = _S_NEWLINE_CHARW;
					WriteFile(cmnGetLogFileHandle(), &c, sizeof(c), (LPDWORD)&w, NULL);	
					//
					// CRLF jupms to next line in file
					//
					// Jump to next char
					continue;
				}
				// Write non-CRLF char
				WriteFile(cmnGetLogFileHandle(), &c, sizeof(c), (LPDWORD)&w, NULL);				
			}
		}
		else
		{
			// Has to convert to ANSI first
			char szOut[512];
			DWORD dwSize = 512;

			cmnConvertUNICODEToANSI(szOut, &dwSize, wszOut);

			// Now write as if it is ANSI
			char c;

			for(int i = 0; i < (int)strlen(szOut); i++)
			{
				//
				// Get char from output buffer
				c = szOut[i];
			
				//
				// Detects CRLF chars
				if(c == _S_NEWLINE_CHARA) {

					// Write CR
					c = _S_CR_CHARA;
					WriteFile(cmnGetLogFileHandle(), &c, sizeof(c), (LPDWORD)&w, NULL);

					// Write LF
					c = _S_NEWLINE_CHARA;
					WriteFile(cmnGetLogFileHandle(), &c, sizeof(c), (LPDWORD)&w, NULL);	
					//
					// CRLF jupms to next line in file
					//
					// Jump to next char
					continue;
				}
				// Write non-CRLF char
				WriteFile(cmnGetLogFileHandle(), &c, sizeof(c), (LPDWORD)&w, NULL);
			}
		}
	}				
	va_end(args);	
}
/*
 *	Writes formatted string to output window and to log file
 *
 */
void cmnTraceA(LPCSTR lpcszFormat, ...)
{
	DWORD w;
	int nBufferSize;
	char szBuffer[512];
	char szOut[512];
	LPCSTR lpFormat;
	BOOL	bPureMsg = FALSE;
	char	szFileName[MAX_PATH];

	// Because the failing filename is stored in UNICODE, then we have to convert the string back to ANSI, so
	// ANSI logging is generally slower. But on the other hand UNICODE logging has to write 2 times more bytes
	// to the logfile
	// Not using dynamic allocation, because this should be faster :)
	char szFailFileName[MAX_PATH];
	// Temporary use w, so we don't have to create another DWORD
	w = MAX_PATH;
	// After the call, the szFailName will contain the name of the file, where the error occured
	cmnConvertUNICODEToANSI(szFailFileName, &w, cmnGetFailFileName());

    va_list args;
	va_start(args, lpcszFormat);

	char * cSemicolon = strchr(lpcszFormat, _S_SEMICOLON_CHARA);
	if(cSemicolon != NULL && cSemicolon == lpcszFormat) {
		lpFormat = "%s\n";
		bPureMsg = TRUE;
	}
	else 
	{
		lpFormat = "* %s[%d]: %s\n";
		bPureMsg = FALSE;
	}

	nBufferSize = _vsnprintf(szBuffer, _COUNTOF(szBuffer), lpcszFormat, args);
	va_end(args);

	if(nBufferSize >= 0) {
		if(bPureMsg) {
			nBufferSize = sprintf(szOut, lpFormat, szBuffer);
		}
		else {
			cmnCutDirectory(szFileName, szFailFileName);
			nBufferSize = sprintf(szOut, lpFormat, szFileName, cmnGetFailLineNumber(), szBuffer);
		}
		
		if(nBufferSize >= 0) {
			OutputDebugString(szOut);
		}
	}

	if(cmnIsLoggingToFile() && cmnGetLogFileHandle())
	{
		if(!cmnIsLogFileUnicode())
		{
			char c;

			for(int i = 0; i < (int)strlen(szOut); i++) {
				//
				// Get char from output buffer
				c = szOut[i];
			
				//
				// Detects CRLF chars
				if(c == _S_NEWLINE_CHARA) {

					// Write CR
					c = _S_CR_CHARA;
					WriteFile(cmnGetLogFileHandle(), &c, sizeof(c), (LPDWORD)&w, NULL);

					// Write LF
					c = _S_NEWLINE_CHARA;
					WriteFile(cmnGetLogFileHandle(), &c, sizeof(c), (LPDWORD)&w, NULL);	
					//
					// CRLF jupms to next line in file
					//
					// Jump to next char
					continue;
				}
				// Write non-CRLF char
				WriteFile(cmnGetLogFileHandle(), &c, sizeof(c), (LPDWORD)&w, NULL);
			}
		}
		else
		{
			// Has to convert to UNICODE first
			WCHAR wszOut[512];
			DWORD dwSize = 512;

			cmnConvertANSIToUNICODE(wszOut, &dwSize, szOut);

			// Now write as if it is UNICODE
			WCHAR c;

			for(int i = 0; i < (int)wcslen(wszOut); i++)
			{
				//
				// Get char from output buffer
				c = wszOut[i];
			
				//
				// Detects CRLF chars
				if(c == _S_NEWLINE_CHARW) {
					
					// Write CR
					c = _S_CR_CHARW;
					WriteFile(cmnGetLogFileHandle(), &c, sizeof(c), (LPDWORD)&w, NULL);
					
					// Write LF
					c = _S_NEWLINE_CHARW;
					WriteFile(cmnGetLogFileHandle(), &c, sizeof(c), (LPDWORD)&w, NULL);	
					//
					// CRLF jupms to next line in file
					//
					// Jump to next char
					continue;
				}
				// Write non-CRLF char
				WriteFile(cmnGetLogFileHandle(), &c, sizeof(c), (LPDWORD)&w, NULL);				
			}
		}
	}
	va_end(args);
}

/*
 *	Writes error message with error code which is analyzed
 */
void		cmnTraceErrW(LPCWSTR lpcwszMsg, HRESULT hr)
{
	LPCWSTR	lpFormat = L"%s > %s";
	wchar_t   wszOut[512];
	wchar_t   wszTemp[512];	
	// Get DX error code
	wcscpy(wszTemp, DXGetErrorString8W(hr));
	//
	if(wcscmp(wszTemp, UNKNOWN_DX_ERRW)) {
		swprintf(wszOut, lpFormat, lpcwszMsg, wszTemp);		
	}	
	//
	// If nor DX error code is valid, try to get Win32 API error code
	else {
		LPVOID lpMsgBuf;
		FormatMessageW(FORMAT_MESSAGE_ALLOCATE_BUFFER | 
					FORMAT_MESSAGE_FROM_SYSTEM | 
					FORMAT_MESSAGE_IGNORE_INSERTS,
					NULL,hr,
					MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), 
					(LPWSTR) &lpMsgBuf,	0,	NULL);

		swprintf(wszOut, lpFormat, lpcwszMsg, (LPCWSTR)lpMsgBuf);					
		LocalFree(lpMsgBuf);
	}
	
	cmnTraceW(wszOut);	
}

void		cmnTraceErrA(LPCSTR lpcszMsg, HRESULT hr)
{
	LPCSTR	lpFormat = _T("%s > %s");
	char   szOut[512];
	char   szTemp[512];	
	// Get DX error code
	strcpy(szTemp, DXGetErrorString8A(hr));
	//
	if(strcmp(szTemp, UNKNOWN_DX_ERRA)) {
		sprintf(szOut, lpFormat, lpcszMsg, szTemp);		
	}	
	//
	// If nor DX error code is valid, try to get Win32 API error code
	else {
		LPVOID lpMsgBuf;
		FormatMessageA(FORMAT_MESSAGE_ALLOCATE_BUFFER | 
					FORMAT_MESSAGE_FROM_SYSTEM | 
					FORMAT_MESSAGE_IGNORE_INSERTS,
					NULL,hr,
					MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), 
					(LPSTR) &lpMsgBuf,	0,	NULL);

		sprintf(szOut, lpFormat, lpcszMsg, (LPCSTR)lpMsgBuf);					
		LocalFree(lpMsgBuf);
	}
	
	cmnTraceA(szOut);
}

void	cmnThrowW(LPCWSTR lpcwszFormat, ...)
{
	int nBufferSize;
	wchar_t wszBuffer[512];	
	wchar_t wszOut[512];
	LPCWSTR lpFormat = L"* Exception * %s";

	va_list args;
	va_start(args, lpcwszFormat);	

	nBufferSize = _vsnwprintf(wszBuffer, _COUNTOF(wszBuffer), lpcwszFormat, args);

	va_end(args);
	
	if(nBufferSize >= 0) { 
		
		nBufferSize = swprintf(wszOut, lpFormat, wszBuffer);		
	}		
	//
	// Throw string
	if(nBufferSize >= 0) {
		throw  wszOut;
	}
}

void	cmnThrowA(LPCSTR lpcszFormat, ...)
{
	int nBufferSize;
	char szBuffer[512];	
	char szOut[512];
	LPCSTR lpFormat = "* Exception * %s";

	va_list args;
	va_start(args, lpcszFormat);	

	nBufferSize = _vsnprintf(szBuffer, _COUNTOF(szBuffer), lpcszFormat, args);

	va_end(args);
	
	if(nBufferSize >= 0) { 
		
		nBufferSize = sprintf(szOut, lpFormat, szBuffer);		
	}		
	//
	// Throw string
	if(nBufferSize >= 0) {
		throw  szOut;
	}
}

void	cmnThrowErrW(LPCWSTR _lpwszMsg, HRESULT hr)
{
	LPCWSTR	lpFormat = L"%s > %s";
	wchar_t   wszOut[512];
	wchar_t   wszTemp[512];	
	//
	// Try to get dx error code
	wcscpy(wszTemp, DXGetErrorString8W(hr));
	//
	if(wcscmp(wszTemp, UNKNOWN_DX_ERRW)) {
		swprintf(wszOut, lpFormat, _lpwszMsg, wszTemp);		
	}	
	//
	// If nor DX error code is valid, try to get Win32 API error code
	else {
		LPVOID lpMsgBuf;
		FormatMessageW(FORMAT_MESSAGE_ALLOCATE_BUFFER | 
					FORMAT_MESSAGE_FROM_SYSTEM | 
					FORMAT_MESSAGE_IGNORE_INSERTS,
					NULL,hr,
					MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), 
					(LPWSTR) &lpMsgBuf,	0,	NULL);
		swprintf(wszOut, lpFormat, _lpwszMsg, (LPCWSTR)lpMsgBuf);					
		LocalFree(lpMsgBuf);
	}
	cmnThrowW(wszOut);	
}

void	cmnThrowErrA(LPCSTR _lpszMsg, HRESULT hr)
{
	LPCSTR	lpFormat = _T("%s > %s");
	char   szOut[512];
	char   szTemp[512];	
	//
	// Try to get dx error code
	strcpy(szTemp, DXGetErrorString8A(hr));
	//
	if(strcmp(szTemp, UNKNOWN_DX_ERRA)) {
		sprintf(szOut, lpFormat, _lpszMsg, szTemp);		
	}	
	//
	// If nor DX error code is valid, try to get Win32 API error code
	else {
		LPVOID lpMsgBuf;
		FormatMessageA(FORMAT_MESSAGE_ALLOCATE_BUFFER | 
					FORMAT_MESSAGE_FROM_SYSTEM | 
					FORMAT_MESSAGE_IGNORE_INSERTS,
					NULL,hr,
					MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), 
					(LPSTR) &lpMsgBuf,	0,	NULL);
		sprintf(szOut, lpFormat, _lpszMsg, (LPCSTR)lpMsgBuf);					
		LocalFree(lpMsgBuf);
	}
	cmnThrowA(szOut);	
}

/*
 *	Shows message box 
 */
void	cmnMessageBoxW(LPCWSTR lpwszCaption, UINT uType)
{	
	LPWSTR lpwszTitle;
	//
	// Find correct title
	switch(uType) {
	case MSG_ERROR:
		lpwszTitle = _S_ERRORW;
		break;
	case MSG_WARNING:
		lpwszTitle = _S_WARNINGW;
		break;
	case MSG_INFO:
		lpwszTitle = _S_INFOW;
		break;
	default:
		lpwszTitle = _S_COMMONW;
		break;
	}
	// 
	// Set additional msg box window styles 
	uType |= (MB_TOPMOST|MB_OK);
	//
	// Show window
	::MessageBoxW(NULL, lpwszCaption, lpwszTitle, uType);
}

void	cmnMessageBoxA(LPCSTR lpszCaption, UINT uType)
{	
	LPSTR lpszTitle;
	//
	// Find correct title
	switch(uType) {
	case MSG_ERROR:
		lpszTitle = _S_ERRORA;
		break;
	case MSG_WARNING:
		lpszTitle = _S_WARNINGA;
		break;
	case MSG_INFO:
		lpszTitle = _S_INFOA;
		break;
	default:
		lpszTitle = _S_COMMONA;
		break;
	}
	// 
	// Set additional msg box window styles 
	uType |= (MB_TOPMOST|MB_OK);
	//
	// Show window
	::MessageBox(NULL, lpszCaption, lpszTitle, uType);
}

 /*
  *	This func is called when user uses macros TRACE or THROW
  * to save file name and line in the source file
  *
  * System needs this data because displays file and line 
  */
COMMON_API void	cmnSavePosW(LPCWSTR _lpcwszFile, DWORD _dwLine)
{
	if(_lpcwszFile)
	{
		wcscpy(g_sLogSettings.wszFailFileName, _lpcwszFile);
	}

	g_sLogSettings.dwFailLine = _dwLine;
}

COMMON_API void cmnSavePosA(LPCSTR _lpcszFile, DWORD _dwLine)
{
	WCHAR wszFile[MAX_PATH];	// Has to be enough
	DWORD dwSize = MAX_PATH;
	
	cmnConvertANSIToUNICODE(wszFile, &dwSize, _lpcszFile);	// Auto-detect length

	// Now call the UNICODE version
	cmnSavePosW(wszFile, _dwLine);
}

// DBGTRACE implementation
#ifdef _DEBUG

	// Don't need this
	//#include <stdio.h>
	//#include <stdarg.h>

	// Writes the output to the debugger window in case of debug build, in release version the
	// macro TRACE is (void)0 call, resulting in no call
	// WARNING : The function cannot handle when the format string contains the % sign preceded by
	//			 the : char, the string then ends after the : char. This bug is also contained in
	//			 the MicroSoft's MFC TRACE macro

	//#pragma COMMENT("FIXED --- The macro TRACE cannot write floating point values --- FIXED")
	//#pragma COMMENT("For tracing in this module (i.e. Common.DLL) always use DBGTRACEA macro")

	COMMON_API void dbgDebugTraceW(LPCWSTR lpszFormat, ...)
	{
		va_list args;
		va_start(args, lpszFormat);

		int nBufferSize;
		WCHAR szBuffer[512];	// The message must have length < 512

		nBufferSize = _vsnwprintf(szBuffer, _COUNTOF(szBuffer), lpszFormat, args);	// This makes the buffer

		if(nBufferSize >= 0)
		{
			OutputDebugStringW(szBuffer);
		}

		va_end(args);
	}

	COMMON_API void dbgDebugTraceA(LPCSTR lpszFormat, ...)
	{
		va_list args;
		va_start(args, lpszFormat);

		int nBufferSize;
		TCHAR szBuffer[512];	// The message must have length < 512

		nBufferSize = _vsnprintf(szBuffer, _COUNTOF(szBuffer), lpszFormat, args);	// This makes the buffer

		if(nBufferSize >= 0)
		{
			OutputDebugStringA(szBuffer);
		}

		va_end(args);
	}

#endif
